import numpy as np

def savitzky_golay(y, window_size=10, order=3, deriv=0, rate=1):
	from math import factorial

	window_size = np.abs(np.int(window_size))
	order = np.abs(np.int(order))

	order_range = range(order+1)
	half_window = (window_size -1) // 2
	# precompute coefficients
	b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
	m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
	# pad the signal at the extremes with
	# values taken from the signal itself
	firstvals = y[0] - np.abs( y[1:half_window+1][::-1] - y[0] )
	lastvals = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
	y = np.concatenate((firstvals, y, lastvals))
	return np.convolve( m[::-1], y, mode='valid')



import h5py, sys
import numpy as np
import argparse

def sub(parser):
	parser.add_argument('index', help='the index')
	parser.add_argument('input', help='the hdf5 methylation file')
	parser.add_argument('cpgs', help='the cpg hdf5 file')


def run(args):
	f = h5py.File(args.input)
	index = h5py.File(args.index)
	fc = h5py.File(args.cpgs)

	all_cov = 0
	all_length = 0

	for chrom in map(str, range(1,23)):
		arr = f[chrom]['methylation']
		arr_cpg = fc[chrom]["positions"]
		arr_ind = index[chrom]["cpg_positions"]

		cpg_ind = np.searchsorted(arr_ind, arr_cpg)

		for start, stop in cpg_ind:
			cpg_islang_cov = np.sum(arr[start:stop])
			all_cov += cpg_islang_cov
			all_length += stop-start

	print(all_cov / all_length)


#		print(chrom, round(all_sum / arr.shape[0], 2), sep='\t')


if __name__ == "__main__":
	parser = argparse.ArgumentParser()
	sub(parser)
	args = parser.parse_args()
	run(args)



import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import seaborn as sns
import os, inspect
import matplotlib.gridspec as gridspec
import scipy.spatial.distance as distance
import scipy.cluster.hierarchy as sch
import numpy as np
import h5py
import argparse
from matplotlib.font_manager import FontProperties

from functools import reduce

from camel.wrap.sample import Sample
from camel.helper.common import savitzky_golay

from sklearn.decomposition import PCA

sns.set_style("whitegrid")

def gen_data(samples):
    sample_chromosomes = set(samples[0].chromosomes())
    chromosomes = [c for c in map(str, range(1,100)) if c in sample_chromosomes]

    M = []

    for chrom in chromosomes:
        print("loading chromosome {}".format(chrom))
        coverages = [s.coverage(chrom) for s in samples]

        for c in coverages:
            c[c==0] = -1

        val = reduce(np.logical_and, [c > 10 for c in coverages])

        M.append(np.array([s.m_values(chrom, smooth_window=10) for s in samples]))

    return np.hstack(M)

#    sd = np.std(M, axis=0) * val
#    sorted_indices = np.argsort(sd)[::-1]
#    return M.T[sorted_indices[:size]]


def plot(samples, labels=[], colors=[], markers=[]):
    '''return pca plot figure for samples'''
    # create a new figure
    fig = plt.figure(figsize=(10, 10))

    # permute test data and make dataframe
    data = gen_data(samples)

    print("received data, calculating plot")

    pca = PCA(n_components=2)
    proj = pca.fit_transform(data)
    #PCA(copy=True, n_components=2, whiten=False)
#    print("explained_variance_ratio_", pca.explained_variance_ratio_)

    if not colors:
        colors = ['black'] * len(samples)

    if not markers:
        markers = ['o'] * len(samples)


    if not labels:
        labels = samples

    fig = plt.figure()


    g = sns.pointplot(proj[:,0], proj[:,1], color="#336699", marker='o', join=False, dodge=False)
    g.cla()

    #g.plot(0, 0, color="red", marker="o", ms=0, label="test")

    for x, y, c, m, l in zip(proj[:,0], proj[:,1], colors, markers, labels):
        g.plot(x, y, color=c, alpha=0.6, marker=m, ms=13, label=l, linestyle="")

    explained_variance_ratio = pca.explained_variance_ratio_

    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.xlabel("PC1 ({0:.2f} explained variance)".format(explained_variance_ratio[0]))
    plt.ylabel("PC2 ({0:.2f} explained variance)".format(explained_variance_ratio[1]))

    return fig


def run(args):
    samples = [Sample(s) for s in args.samples]
    fig = plot(samples, labels=args.labels, colors=args.colors, markers=args.markers)
    fig.savefig(args.output, bbox_inches='tight')


def sub(parser):
    parser.add_argument("output", help="the output file for the plot")
    parser.add_argument("--samples", nargs="+", help="methylation values in camel format")
    parser.add_argument("--labels", nargs="*", help="the labels of the samples for the clustering plot")
    parser.add_argument("--colors", nargs="*", help="the colors of the markers")
    parser.add_argument("--markers", nargs="*", help="the style of the markers, use matplotlib maker symbols")


def main():
    parser = argparse.ArgumentParser(description='create a clustering plot')
    sub(parser)
    args = parser.parse_args()
    run(args)


if __name__ == "__main__":
    main()



#call all meth informations for each cpg

import pysam
import time
import sys
import h5py
import argparse
import numpy as np
import math
import os

from bisect import bisect_left
from collections import defaultdict, namedtuple
from itertools import islice

Hit = namedtuple("Hit", "typ, rel_pos, qual, index")

def get_positions(left, right, gc):
    #perform a bidirectional search for gc positions
    pos = bisect_left(gc, left)
#    pos = np.searchsorted(gc, left)

    while pos < len(gc) and gc[pos] <= right:
        yield (gc[pos], pos)
        pos += 1

def get_hits(read, positions):
    #ignore last position on forward and last position on reverse
    hit_positions = get_positions(read.pos + 1, read.aend - 1, positions[read.tid])
    seq = read.seq.upper()

    for pos, index in hit_positions:
        rel_pos = relative_pos(read, pos-1)

        if (rel_pos == None):
            continue

        pair_orientation = read.is_read1 == read.is_reverse

        rel_pos += pair_orientation
        rel_next_pos =  rel_pos + 1 - pair_orientation * 2

        try:
            char = seq[rel_pos]
            next_char = seq[rel_next_pos]
        except:
            print("error", read.qname, read.pos, read.cigar, hit_positions, file=sys.stderr)
            continue

        if pair_orientation and char == 'A' and next_char == 'C':
            typ = 3
        elif not pair_orientation and char == 'C' and next_char == 'G':
            typ = 0
        elif pair_orientation and char == 'G' and next_char == 'C':
            typ = 2
        elif not pair_orientation and char == 'T' and next_char == 'G':
            typ = 1
        elif pair_orientation and next_char == "T":
            typ = 5
        elif not pair_orientation and next_char == "A":
            typ = 4
        else:
            # SNPs
            typ = -1

        ref = "C" if not pair_orientation else "G"

        if rel_pos >= 0:
            yield Hit(typ, rel_pos, ord(read.qual[rel_pos]), index)


def get_hits_nome(read, positions):
    #ignore last position on forward and last position on reverse
    hit_positions = get_positions(read.pos + 1, read.aend - 1, positions[read.tid])

    seq = read.seq.upper()

    for pos, index in hit_positions:
        rel_pos = relative_pos(read, pos-1)

        if (rel_pos == None):
            continue

        pair_orientation = read.is_read1 == read.is_reverse

        rel_pos += 1 - pair_orientation #dirty hack

        try:
            char = seq[rel_pos]
        except:
            print("error", read.qname, read.pos, read.cigar, hit_positions, file=sys.stderr)
            continue

        if not pair_orientation and char == 'T':
            typ = 3
        elif pair_orientation and char == 'G':
            typ = 0
        elif not pair_orientation and char == 'C':
            typ = 2
        elif pair_orientation and char == 'A':
            typ = 1
        else:
            # SNPs
            typ = -1

        if rel_pos >= 0:
            yield Hit(typ, rel_pos, ord(read.qual[rel_pos]), index)


def check_files_exists(*filenames):
    # check filenames
    for f in filenames:
        if not os.path.isfile(f):
            raise Exception("{} not found".format(f))


def run(args):
    bam_filename = args.input
    index_filename = args.index
    output_filename = args.output
    mbq = args.mbq
    mmq = args.mmq
    max_snp_fraction = args.max_snp_fraction


    check_files_exists(index_filename, bam_filename)

    samfile = pysam.Samfile(bam_filename, "r")

    print("load index", file=sys.stderr)
    index = h5py.File(index_filename, 'r')

    print("create datastructures", file=sys.stderr)

    #check if cpg_positions exist / correct index
    for r in index:
        if not "cpg_positions" in index[r]:
            raise Exeption("cpg_positions missing in {}. Please call camel index to create a valid camel-index-file".format(index_filename))
        if args.nome and not "gpc_positions" in index[r]:
            raise Exeption("cpg_positions missing in {}. Please call camel index with --nome param to create a valid camel-index-file for nome calling".format(index_filename))


    # init datastructures
    cpg_positions = [index[r]["cpg_positions"][:].tolist() if r in index else None for r in samfile.references]
    cpg_ctga = [np.zeros((len(index[r]["cpg_positions"]),6), dtype=np.int) if r in index else None for r in samfile.references]


    # init nome datastructures
    gpc_ctga = None
    if args.nome:
        gpc_positions = [index[r]["gpc_positions"][:].tolist() if r in index else None for r in samfile.references]
        gpc_ctga = [np.zeros((len(index[r]["gpc_positions"]),6), dtype=np.int) if r in index else None for r in samfile.references]

#    imprinting_count = dict((samfile.gettid(c), np.zeros((len(index[c]["cpg_positions"]),4), dtype=np.int)) for c in chromosomes)

#    per_pos_stats = [[]] * 101
#    for i in range(len(per_pos_stats)):
#        per_pos_stats[i] = [0,0]

#    count = 0
    timestart = time.time() #measure running time

    print("open samfile", file=sys.stderr)

    samout = args.samout
    ignore_duplicates = args.ignore_duplicates

    if samout:
        #pipethrough header
        print("sam output activated", file=sys.stderr)
        outfile = pysam.AlignmentFile("-", "wh", header=samfile.header)

    if ignore_duplicates:
        from camel.helper.bloomfilter import Bloomfilter, calculate_bloomfilter_size
        # calculate bloomfilter params
        bitcount, hash_function_count = calculate_bloomfilter_size(args.read_count, args.error_rate)
        print("create bloomfilter for duplicate filtering ({bitcount} bits, {hashcount} hash functions)".format(bitcount=bitcount, hashcount=hash_function_count), file=sys.stderr)
        bf = Bloomfilter(bitcount, hash_function_count)

    camel_duplicates = set()
    mark_duplicates = set()

    for count, read in enumerate(samfile):
        if samout:
            #pipethrough reads
            outfile.write(read)

        if count % 1000000 == 0:
            print(count, "reads processed", file=sys.stderr)
#            if count == 1000000: break

        # filter out duplicates and unmapped reads

        if read.is_duplicate or read.mapq < mmq or not cpg_positions[read.tid] or read.is_unmapped:
            continue

        def read_key(read):
            '''return the positioning key for an read used for duplication detection'''
            if read.is_reverse:
                return (read.next_reference_id, read.next_reference_start, read.isize)
            else:
                return (read.reference_id, read.reference_start, read.isize)

        def check_dup():
            pos = read_key(read)
            is_seen = bf.lookup(pos)
            bf.add(pos)
            return is_seen

        is_duplicate = None

        #TODO: use read.get_reference_positions?

        # for each cpg position on the read
        for h in get_hits(read, cpg_positions):
            #perform duplication check (only if there are hits
            if ignore_duplicates and is_duplicate is None:
                is_duplicate = check_dup()
            if is_duplicate:
                break

            # add only if minumum quality is reached
            if h.qual < mbq or h.typ == -1: continue
            # handle duplicate checks
            cpg_ctga[read.tid][h.index, h.typ] += 1

        # nome calling
        if args.nome:
            #perform duplication check (only if there are hits
            if ignore_duplicates and is_duplicate is None:
                is_duplicate = check_dup()
            if is_duplicate:
                break

            # for each gpc position on the read
            for h in get_hits_nome(read, gpc_positions):
                # add only if minumum quality is reached
                if h.qual < mbq or h.typ == -1: continue
                gpc_ctga[read.tid][h.index, h.typ] += 1

    if samout:
        outfile.close()


    #            if prev_h:
    #                imprinting_typ = ((h.typ & 1) << 1) + (prev_h.typ & 1)
    #                imprinting_count[read.tid][h.index, imprinting_typ] += 1
    #            prev_h = h


    def remove_snps(values_ctga, max_snp_fraction=0.2):
        for x in values_ctga:
            if x is None:
                continue
            first_coverage = np.sum(x[:,(2,3,5)], axis=1)
            first_coverage[first_coverage==0] = -1
            first_ratio = x[:,5] / first_coverage

            second_coverage = np.sum(x[:,(0,1,4)], axis=1)
            second_coverage[second_coverage==0] = -1
            second_ratio = x[:,4] / second_coverage

            select = (first_ratio > max_snp_fraction) & (x[:,5] >= 2) | (second_ratio > max_snp_fraction) & (x[:,4] >= 2)
            rselect = np.repeat(select, 6).reshape((-1,6))
            x[rselect] = 0


    print("removing c->t snps", file=sys.stderr)
    remove_snps(cpg_ctga, max_snp_fraction)

    print("removing g->a snps (NOMe)", file=sys.stderr)
    if args.nome:
        remove_snps(gpc_ctga, max_snp_fraction)

    print(time.time() - timestart, "seconds for calling", file=sys.stderr)


    # saving all data
    save(output_filename, samfile, cpg_ctga, gpc_ctga) #imprinting_count=imprinting_count

    print(time.time() - timestart, "seconds", file=sys.stderr)


def save(output_filename, samfile, cpg_ctga, gpc_ctga, imprinting_count=None):
    #save hdf5 file
    out_file = h5py.File(output_filename, 'w')

    #writing cpg_positions
    for tid in range(len(cpg_ctga)):
        if cpg_ctga[tid] is None: continue
        chrom       = samfile.getrname(tid)
        grp         = out_file.create_group(chrom)
        data_methyl = cpg_ctga[tid][:,0:4]
        print("writing", samfile.getrname(tid), data_methyl.shape, file=sys.stderr)
        dset        = grp.create_dataset("methylation", dtype='i', data=data_methyl, maxshape=data_methyl.shape)

        if imprinting_count:
            data_imprinting = imprinting_count[tid]
            dset = grp.create_dataset("imprinting", dtype='i', data=data_imprinting, maxshape=data_imprinting.shape)

    if gpc_ctga:
        #writing gpc_positions
        for tid in range(len(gpc_ctga)):
            if gpc_ctga[tid] is None: continue
            chrom       = samfile.getrname(tid)
            grp         = out_file[chrom]
            data_methyl = gpc_ctga[tid][:,0:4]
            print("writing", samfile.getrname(tid), data_methyl.shape, file=sys.stderr)
            dset        = grp.create_dataset("nome", dtype='i', data=data_methyl, maxshape=data_methyl.shape)

    out_file.close()


def relative_pos(read, pos):
    # calculate the position of the base, without deletions and insertions
    relative_pos = pos - read.pos
    cigar        = read.cigar
    curr_pos     = 0

    for c in cigar:
        if curr_pos + c[1] <= relative_pos:
            if c[0] == 2 or c[0] == 4: #deletion or soft clipping
                relative_pos -= c[1]
            if c[0] == 1: #insertion
                relative_pos += c[1]
            else: #match
                curr_pos += c[1]
        else:
            if c[0] == 2: #cpg inside deletion
                return None
            else:
                return relative_pos
    return None


def sub(parser):
    parser.add_argument('input',help='The input bam file.')
    parser.add_argument('index', help='The camel-index-file.')
    parser.add_argument('output', help='The output camel-data file.')
    parser.add_argument('-mbq', nargs='?', help='minimum pred-scaled base quality score', default=17, type=int)
    parser.add_argument('-mmq', nargs='?', help='minimum pred-scaled mapping quality score', default=30, type=int)
    parser.add_argument('--nome', '-n', action='store_true', help='perform an additional nome calling')
    parser.add_argument('--samout', '-s', action='store_true', help='print sam to stdout')
    parser.add_argument('--ignore-duplicates', '-d', action='store_true', help='Activate the ignore-duplication mode to perform own pcr-duplication detection by read positions using a bloomfilter. The setup of the bloomfilter requires an aproximate read-count and an error-rate. Note that the default value requires 3.35GB of memory.')
    parser.add_argument('--read-count', '-r', metavar="N", nargs='?', help='Set the input read count, used for size determination of the bloomfilter (default: 1000000000). Only needed in ignore-duplication mode', default=1000000000, type=int)
    parser.add_argument('--error-rate', '-e', metavar="N", nargs='?', help='Set the error-rate, used for size determination of the bloomfilter (default: 0.00001). Only needed in ignore-duplication mode.', default=0.0001, type=float)
    parser.add_argument('--max_snp_fraction', "-f", help="Set the maximum fraction of basepairs to not discard a CpG as SNP. (CpGs with a snp fraction higher than this values get discarded)", default=0.2, type=int)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Calculate methylation level for each CpG.')
    sub(parser)
    args = parser.parse_args()
    run(args)



import argparse
import numpy as np
from camel.wrap.region import Region


def run(args):
    out(args)


def run_direct(dmrs_filename, track_filename, name="ann", type="overlap", n=1):
    region = Region(track_filename)

    for line in open(dmrs_filename):
        line = line.strip()

        if line.startswith('#'):
            print(line.strip(), name + ".count", sep='\t')
            continue

        if not line:
            continue

        split = line.split('\t')
        chrom, start, stop = split[:3]

        if type=="overlap":
            annotation_values, typs, dists = region.overlap(chrom.encode(), start, stop)
            yield (line, annotation_values, typs, dists)
        elif type=="nearest":
            annotation_values, typs, dists = region.nearest(chrom.encode(), start, stop, n)
            yield (line, annotation_values, typs, dists)
        elif type=="neighbours":
            annotation_values, typs, dists = region.neighbours(chrom.encode(), start, stop)
            yield (line, annotation_values, typs, dists)



def out(args):
    if args.type == "overlap":
        ana_type="overlap"
    elif args.type == "overlap-count":
        ana_type="overlap"
    elif args.type == "nearest":
        ana_type="nearest"
    elif args.type == "neighbours":
        ana_type="neighbours"

    data = run_direct(args.dmrs, args.track, args.name, ana_type, args.number)

    for dmr, annotation, typ, dist in data:
        print(dmr, end="\t")
        if args.type == "overlap":
            names = [a["name"].decode() for a in annotation]
            print(*names, sep=",")
        elif args.type == "overlap-count":
            print(len(annotation))
        elif args.type == "nearest":
            names = [a["name"].decode() for a in annotation]
            distances = [dists for a in annotation]
            print(*names, sep=",", end="\t")
            print(*distances, sep=",")
        elif args.type == "neighbours":
            names = [a["name"].decode() for a in annotation]
            distances = [dists for a in annotation]
            print(*names, sep=",", end="\t")
            print(*distances, sep=",")

#        elif len(annotation) == 0:
#            print("\t")
#        else:
#            print(*[a["name"].decode() for a in annotation], sep=",")


def sub(parser):
    parser.add_argument('dmrs', help="the track with the DMRs")
    parser.add_argument('track', help="the track with the annotation data")
    parser.add_argument('name', help="the prefix-name of the new annotation column")
    parser.add_argument('--type', '-t', choices=["overlap", "overlap-count", "nearest", "neighbours"], default="overlap")
    parser.add_argument('--number', '-n', default=1, type=int)


def main():
    parser = argparse.ArgumentParser(description='Annotate DMRs by track')
    sub(parser)
    args = parser.parse_args()
    out(args)


if __name__ == '__main__':
    main()



import argparse
import os, inspect

# add parent parent path to system path
if __name__ == "__main__":
    currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    parentdir = os.path.dirname(os.path.dirname(currentdir))
    os.sys.path.insert(0,parentdir)

from camel.modules import call, view, merge, index, statistics, meta, dmr, cluster, pca, annotate, average

def run():
    p = argparse.ArgumentParser(description='Camel')
    subparsers = p.add_subparsers(dest="subcommand")

    # adding alle subparser params

    subp = subparsers.add_parser("call", help="calculate methylation level for each CpG by mapped bam file")
    call.sub(subp)

    subp = subparsers.add_parser("view", help="view the methylation information")
    view.sub(subp)

    subp = subparsers.add_parser("dmr", help="compute differentially methylated regions")
    dmr.sub(subp)

    subp = subparsers.add_parser("merge", help="merge samples")
    merge.sub(subp)

    subp = subparsers.add_parser("index", help="create the index for calling")
    index.sub(subp)

    subp = subparsers.add_parser("statistics", help="compute statistical informations")
    statistics.sub(subp)

    subp = subparsers.add_parser("meta", help="access the meta informations")
    meta.sub(subp)

    subp = subparsers.add_parser("cluster", help="create a clustering plot")
    cluster.sub(subp)

    subp = subparsers.add_parser("pca", help="create a principle component analysis plot")
    pca.sub(subp)

    subp = subparsers.add_parser("annotate", help="annotate a dmr file")
    annotate.sub(subp)

    subp = subparsers.add_parser("average", help="view average methylation of specified regions")
    average.sub(subp)

    args = p.parse_args()

    if not args.subcommand:
        print(p.format_help())
        return

    switch = {
        "call" : call,
        "view" : view,
        "average" : average,
        "index" : index,
        "statistics" : statistics,
        "meta" : meta,
        "cluster" : cluster,
        "pca" : pca,
        "annotate" : annotate,
        "merge" : merge,
        "dmr": dmr,
    }

    # calling the correct submodule
    switch[args.subcommand].run(args)

if __name__ == '__main__':
    run()



import matplotlib
matplotlib.use('Agg')
import h5py
import numpy as np
from collections import defaultdict
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import argparse
import os

# add parent parent path to system path
#if __name__ == "__main__":
#    currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
#    parentdir = os.path.dirname(os.path.dirname(currentdir))
#    os.sys.path.insert(0,parentdir)

from camel.wrap.region import Region
from camel.wrap.sample import Sample


def savitzky_golay(y, window_size, order, deriv=0, rate=1):
    import numpy as np
    from math import factorial

    window_size = np.abs(np.int(window_size))
    order = np.abs(np.int(order))

    order_range = range(order+1)
    half_window = (window_size -1) // 2
    # precompute coefficients
    b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
    m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs( y[1:half_window+1][::-1] - y[0] )
    lastvals = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve( m[::-1], y, mode='valid')


def run(args):
    fig = plot(args.sample, args.index, args.genes, args.cgi, args.incgi, args.undirected)
    fig.savefig(args.output)


def plot(sample_filename, index_filename, genes_filename, cgi_filename, in_cpg_islands=True, undirected=False):
    fig = plt.figure(frameon=True)

    # open transcription start sites bed file:

    genes = Region(genes_filename)
    cpg_islands = Region(cgi_filename)
    sample = Sample(sample_filename)
    index = h5py.File(index_filename)

    size = 1000

    vector = np.zeros(size * 2, dtype=np.float)
    vector_cov = np.zeros(size * 2, dtype=np.float)

    for chrom in sample.chromosomes(): # for each chromosome
        if chrom not in list(map(str,range(1,23))) + ["X", "Y"]:
            continue

        if not chrom in index: # exept for these, that are not in the index
            continue

        cpg_islands_valid = cpg_islands.values["chrom"] == chrom.encode()
        cpg_islands_starts = cpg_islands.values[cpg_islands_valid]["start"]
        cpg_islands_stops = cpg_islands.values[cpg_islands_valid]["stop"]
        order = np.argsort(cpg_islands_starts)
        cpg_islands_starts = cpg_islands_starts[order]
        cpg_islands_stops  = cpg_islands_stops[order]

        value_matrix = sample.methylation(chrom, nome=True) # nome methylation values
        gpc_positions = index[chrom]["gpc_positions"][:] # get gpc positions out of the index file
        gpc_valid = index[chrom]["valid_gpc"][:]

        genes_valid = genes.values["chrom"] == chrom.encode()
        genes_values = genes.values[genes_valid]

        tss_positions = genes_values["stop"]
        directions = genes_values["direction"]
        tss_positions[directions] = genes_values["start"][directions]

        order = np.argsort(tss_positions)
        tss_positions = tss_positions[order]
        directions = directions[order]

        in_cpg_island = np.searchsorted(cpg_islands_starts,tss_positions) == np.searchsorted(cpg_islands_stops, tss_positions) + 1

        #reverse cpg valid
        if not in_cpg_islands:
            in_cpg_island = np.negative(in_cpg_island) # not in cpg_island

        tss_positions = tss_positions[in_cpg_island]
        directions = directions[in_cpg_island]

        tss_indices = np.searchsorted(gpc_positions, tss_positions)
        tss_indices_minus_1000 = np.searchsorted(gpc_positions, tss_positions - size)
        tss_indices_plus_1000 = np.searchsorted(gpc_positions, tss_positions + size)

#        print(chrom, np.sum(tss_positions), np.sum(tss_indices_minus_1000), np.sum(tss_indices_plus_1000))

        for i in range(len(tss_positions)):
            if tss_indices_minus_1000[i] == tss_indices_plus_1000[i]:
                continue

            tss_position = tss_positions[i]
            positions = gpc_positions[tss_indices_minus_1000[i]:tss_indices_plus_1000[i]]
            values = value_matrix[tss_indices_minus_1000[i]:tss_indices_plus_1000[i]]

            # correct invalid gpc
            gpc_valid_region = gpc_valid[tss_indices_minus_1000[i]:tss_indices_plus_1000[i]]

            corr_values = np.copy(values).reshape((-1,2))
            corr_values = corr_values * gpc_valid_region.reshape((-1,1))
            corr_values = corr_values.reshape((-1,4))

            meth = np.sum(corr_values[:,(0,2)], axis=1)
            cov = np.sum(corr_values, axis=1)

            cov[cov == 0] = -1
            m = meth / cov

            rel_positions = positions - tss_position + size #relative position to tss
            if not undirected:
                rel_positions *= (-1 + directions[i] * 2)

            vector[rel_positions] += m
            vector_cov[rel_positions] += (cov > 0)

    data = 1 - vector / vector_cov # building the average values


    plt.axvline(0, color='#555555')
    plt.xlabel("Distance from TSS (bp)")
    plt.ylabel("1-GpC Methylation (Nuclesome Protection)")
    plt.ylim([0.0, 1.0])
    plt.plot(np.arange(-1000,1000), data, color='#bbbbbb')
    plt.plot(np.arange(-1000,1000),savitzky_golay(data, 50, 3))
    return fig


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('index', help="The camel-index-file.")
    parser.add_argument("genes", help="List of genes in bed format.")
    parser.add_argument("cgi", help="List of cpg-islands in bed format.")
    parser.add_argument("sample", help="A sample file in camel-data format.")
    parser.add_argument("output", help="The output file. The format is chosen by the ending of this file. All matplotlib plot file formats are supported, such as pdf, svg, png or jpg.")
    parser.add_argument("-c", "--incgi", action="store_true")
    parser.add_argument("-u", "--undirected", action="store_true")

    args = parser.parse_args()
    run(args)


if __name__ == '__main__':
    main()



import numpy as np
import h5py
from collections import namedtuple
from os.path import basename, splitext

from camel.helper.common import savitzky_golay

class Sample:
    def __init__(self, filename):
        self.filename = filename
        self.basename = basename(splitext(filename)[0])
        self.f = h5py.File(filename, 'r')

        self.duplication_rate = self.value("meta/duplication_percentage")
        self.lambda_meth = self.value("meta/lambda_meth")
        self.all_meth = self.value("meta/all_meth")
        self.examined_pairs = self.value("meta/examined_pairs")
        self.unpaired_reads = self.value("meta/unpaired_reads")
        self.unmapped = self.value("meta/unmapped_reads")
        self.lane_count = self.value("meta/lanecount")
        self.convertion_rate = 1 - self.lambda_meth if self.lambda_meth is not None else None
        self.read_coverage = self.value("meta/coverage")


        #TODO: check if unmapped_reads are already counted in examined_reads by markduplicates
        if self.unmapped and self.examined_pairs and self.unpaired_reads:
            self.efficiency = 1 - self.unmapped / (self.examined_pairs * 2 + self.unpaired_reads + self.unmapped)
        else:
            self.efficiency = None

    def chromosomes(self):
        return [c for c in self.f]


    def value(self, path):
        if path in self.f:
            return self.f[path][0]
        return None


    def methylation(self, chrom, start_index=None, stop_index=None, nome=False):

        if not nome:
            values = self.f[chrom]["methylation"]
        else:
            values = self.f[chrom]["nome"]

        if start_index and stop_index:
            return values[start_index:stop_index]

        return values[:]


    def _methylated(self, value_matrix):
        '''return the count of the methylated reads of a value matrix'''
        return np.sum(value_matrix[:,(0,2)], axis=1)


    def _coverage(self, value_matrix):
        '''return the coverage of a value matrix'''
        return np.sum(value_matrix, axis=1)


    def methylated(self, chrom, start_index=None, stop_index=None):
        return self._methylated(self.methylation(chrom))


    def coverage(self, chrom, start_index=None, stop_index=None):
        return self._coverage(self.methylation(chrom))


    def m_values(self, chrom, start_index=None, stop_index=None, min_cov=0, smooth_window=0):
        values = self.methylation(chrom, start_index, stop_index)
        coverage = self._coverage(values)
        methylated = self._methylated(values)

        coverage[coverage == 0] = 1 #prevent division by zero afterwards
        m_values = methylated / coverage

        if smooth_window > 0:
            m_values = savitzky_golay(m_values, smooth_window)

        if min_cov > 0:
            m_values = m_values[coverage > min_cov]

        return m_values


    def __repr__(self):
        return self.basename


    def __lt__(self, other):
        return self.basename < other.basename



import numpy as np
import math
import h5py
import os
import sys
import operator
from functools import reduce
import argparse

from numpy.lib.recfunctions import merge_arrays

def load(filename, chrom, min_cov=10):
    f = h5py.File(filename, "r")
    values = f[chrom]["methylation"][:]
    m = np.sum(values[:,(0,2)], axis=1)
    cov = np.sum(values, axis=1)

    cov[cov == 0] = 1
    m_values = m / cov

    return m_values, cov < min_cov


def savitzky_golay(y, window_size, order, deriv=0, rate=1):
    import numpy as np
    from math import factorial

    window_size = np.abs(np.int(window_size))
    order = np.abs(np.int(order))

    order_range = range(order+1)
    half_window = (window_size -1) // 2
    # precompute coefficients
    b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
    m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs( y[1:half_window+1][::-1] - y[0] )
    lastvals = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve( m[::-1], y, mode='valid')


def run(args):
    return run_direct(args.index, args.case, args.control, args.min_cpg, args.min_diff, args.min_cov, args.smooth)


def run_direct(index_filename, case, control, min_cpg=1, min_diff=0, min_cov=10, smooth=False):
    index = h5py.File(index_filename, "r")

    for chrom in map(str, range(1,23)): #range(1,23)
        values_case, invalid_case = zip(*[load(s, chrom, min_cov) for s in case])
        values_control, invalid_control = zip(*[load(s, chrom, min_cov) for s in control])

        #smooth values
#        values_case = list(map(lambda x: savitzky_golay(x, 3, 3), values_case))
#        values_control = list(map(lambda x: savitzky_golay(x, 3, 3), values_control))

        #invalid stored a vector for each position, where at least one coverage is to low
        invalid = reduce(operator.or_, invalid_case + invalid_control)

        #"filter" low coverage
        for value in values_case + values_control:
            value[invalid] = 0

        n_case = len(values_case)
        n_control = len(values_control)

        # calculate means
        mean_case = np.mean(values_case, axis=0)
        mean_control = np.mean(values_control, axis=0)

        # calculate differences and standard deviation
        betas = mean_control - mean_case
        omega = (np.std(values_case + values_control, axis=0) + np.std(values_control, axis=0)) / 2

        #calculate the omega 75th percentile
        omega_perc = np.percentile(omega, 75)

        #cap the omega values to the percentile
        omega[omega >  omega_perc] = omega_perc

        #TODO: smooth the omegas with running mean 101 window

        #form t statistic
        t = np.divide(betas, omega * math.sqrt(1/n_case + 1/n_control))

#        print(t[1873967-4:1873997+10])

#        exit()

        if True:
            t = savitzky_golay(t, 10, 3)

#        print(t[1873967-4:1873997+10])

        t_extend = np.concatenate(([0], t, [0])) # to make sure every dmr has a start and an end


        #TODO: do not delete for further analysis
        #for v, b in zip(*np.histogram(t, range=(-10, 10), bins=100)):
        #    print(b, v, sep="\t")


        over = np.zeros(len(t_extend))
        over[(t_extend > 6) | (t_extend < -6)] = 1
        over_diff = np.diff(over)

        starts = np.where(over_diff==1)[0]
        ends   = np.where(over_diff==-1)[0]

        # the numbers of cpg per dmr
        cpg_count = ends - starts

        # determine the beta value between the dmrs
        cumsum = np.cumsum(betas)


        dmr_betas = np.divide(cumsum[ends - 1] - cumsum[starts - 1], cpg_count) #TODO wahrscheinlich +- 1 auf die indices

        # correct first DMR beta value if the first t-value is significant and the "start" -1 therefore adresses -1, which means the last index
        if starts[0] == 0:
            dmr_betas[0] = cumsum[ends[0] - 1] / cpg_count[0]


        # only dmrs with beta > 0.4 and more than 8 cpg
        selection = (np.abs(dmr_betas) >= min_diff) & (cpg_count >= min_cpg)

        starts_filtered = starts[selection]
        ends_filtered = ends[selection]
        cpg_count = cpg_count[selection]
        dmr_betas = dmr_betas[selection]

        cpg_pos = index[chrom]["cpg_positions"][:]


        m = merge_arrays((np.repeat(chrom, len(starts_filtered)), cpg_pos[starts_filtered], cpg_pos[ends_filtered], cpg_count, dmr_betas), asrecarray=True)
        m.dtype.names = ("chrom", "start", "stop", "n", "diff")

        yield m


def out(x):
    for i in x:
        for a in i:
            print(*a, sep="\t")


def sub(parser):
    # add items to a parser or subparser
    parser.add_argument('index', help='the camel index file')
    parser.add_argument("--case", nargs="+", help="the case samples in camel format")
    parser.add_argument("--control", nargs="+", help="the control samples in camel format")
    parser.add_argument("--min_cpg", default=1, type=int, help="the minimum number of CpGs to call a region")
    parser.add_argument("--min_diff", default=0.0, type=float, help="the minimum mean difference to call a region")
    parser.add_argument("--min_cov", default=10, type=int, help="the minimum coverage for each cpg")
    parser.add_argument("--smooth", action="store_true", help="the methylation varlues get smoothed")



def main():
    parser = argparse.ArgumentParser(description='Calculate methylation level for each CpG.')
    sub(parser)
    args = parser.parse_args()
    out(run(args))


if __name__ == '__main__':
    main()



import sys
import numpy as np
import h5py
from collections import defaultdict, namedtuple
from functools import lru_cache
from numpy.lib.recfunctions import append_fields, rec_append_fields

class Region:
    Reg = namedtuple("Reg", "start stop direction value")


    def __init__(self, filename):
        self.open_txt(filename)


    def clear_chrom(self, chrom):
        if chrom.startswith(b"chr"):
            chrom = chrom[3:]
        return chrom


    def open_txt(self, filename):
        '''open the bed file'''
        with open(filename, 'r') as f:
            for line in f:
                if line.startswith("#"): continue
                break

        column_count = len(line.split("\t"))

        if column_count == 3:
            dtype = [('chrom','a25'),('start','i8'),('stop','a25')]
            converters = {0: lambda x: self.clear_chrom(x)} #TODO: keep that? maybe optional
            values = np.genfromtxt(filename, dtype=dtype, converters=converters)
            names = np.repeat("", len(values))
            directions = np.repeat(True, len(values))
            self.values = append_fields(values, ("name", "direction"), (names, directions))
        elif column_count == 4:
            dtype = [('chrom','a25'),('start','i8'),('stop','i8'),('name','a25')]
            converters = {0: lambda x: self.clear_chrom(x)}
            values = np.genfromtxt(filename, dtype=dtype, converters=converters)
            directions = np.repeat(True, len(values))
            self.values = append_fields(values, ("direction"), (directions))
        elif column_count == 5:
            dtype = [('chrom','a25'),('start','i8'),('stop','i8'),('name','a25'),('score','i8')]
            converters = {0: lambda x: self.clear_chrom(x)}
            values = np.genfromtxt(filename, dtype=dtype, converters=converters)
            directions = np.repeat(True, len(values))
            self.values = append_fields(values, ("direction"), (directions))
        elif column_count == 6:
            dtype = [('chrom','a25'),('start','i8'),('stop','i8'),('name','a25'),('score','i8'),('direction', 'bool')]
            converters = {0: lambda x: self.clear_chrom(x), 5: lambda x: x == b"+"}
            values = np.genfromtxt(filename, dtype=dtype, converters=converters)
            self.values = values
        else:
            sys.exit("invalid number of columns in track")



    def overlap(self, chrom, start, stop):
        '''return list of items that overlap with a given region'''

        chrom = self.clear_chrom(chrom)
        start = int(start)
        stop = int(stop)
        mid = (start + stop) / 2

        valid_chrom = self.values["chrom"] == chrom
        valid_start = self.values["start"] <= stop
        valid_stop  = self.values["stop"] >= start
        valids = valid_chrom & valid_start & valid_stop

        indices = np.nonzero(valids)[0]

        if not len(indices):
            return np.zeros(0, dtype=self.values.dtype), [], []

        valid_values = self.values[indices]
        typs = ((1 + (valid_values["start"] < start)) << 2) + (2 - (valid_values["stop"] > stop))
        dists = (mid - (valid_values["start"] + valid_values["stop"]) / 2).astype(np.int64)

        return valid_values, typs, dists


    def nearest(self, chrom, start, stop, n=1):
        '''return the nearest n regions'''

        chrom = self.clear_chrom(chrom)
        start = int(start)
        stop = int(stop)

        valid_chrom = self.values["chrom"] == chrom
        valid_values = self.values[valid_chrom]

        #calculate distances
        distances_left = start - valid_values["stop"]
        distances_right = valid_values["start"] - stop
        distances = np.fmax(distances_left, distances_right)
        overlaps = distances <= 0
        distances[overlaps] = 0

        # calculate overlap_size
        max_start = np.fmax(start, valid_values["start"])
        min_stop = np.fmin(stop, valid_values["stop"])
        overlap_sizes = min_stop - max_start


#        distances[overlaps] = 0
#        overlap_sizes[np.logical_not(overlaps)] = 0

        sorted_indices = np.argsort(distances)
#        return valid_values[sorted_indices[0:n]]

        distances = distances * (valid_values["direction"] * 2 - 1) * np.sign(distances_left)

        return valid_values[sorted_indices[0:n]], distances[sorted_indices[0:n]], overlap_sizes[sorted_indices[0:n]]


    def neighbours(self, chrom, start, stop, side=True):
        chrom = self.clear_chrom(chrom)
        mid = (int(start) + int(stop)) / 2

        valid_chrom = self.values["chrom"] == chrom
        valid_values = self.values[valid_chrom]

        #calculate distances
        if side:
            ancors = np.copy(valid_values["stop"])
            directions = valid_values["direction"]
            ancors[directions] = valid_values["start"][directions]
        else:
            #use the mids as ancor points
            ancors = (valid_values["start"] + valid_values["stop"]) / 2
        distances = ancors - mid

        left_valid = distances<0
        right_valid = distances>=0

        if np.sum(left_valid):
            left_index = np.argmax(distances[left_valid])
            left_value = valid_values[left_valid][left_index]
            left_distance = distances[left_valid][left_index]
            left_distance = left_distance * (1 - 2 * directions[left_valid][left_index])
        else:
            left_value = np.empty(1, dtype=self.values.dtype)
            left_distance = 0

        if np.sum(right_valid):
            right_index = np.argmin(distances[right_valid])
            right_value = valid_values[right_valid][right_index]
            right_distance = distances[right_valid][right_index]
            right_distance = right_distance * (1 - 2 * directions[right_valid][right_index])
        else:
            right_value = np.empty(1, dtype=self.values.dtype)
            right_distance = 0

        ret_values = np.vstack((left_value, right_value))

        return append_fields(ret_values, ("distance", "overlap"), ((left_distance,right_distance), (0,0)))



import h5py
import numpy as np
import argparse
import sys
from camel.wrap.sample import Sample
from collections import defaultdict
from natsort import natsorted

def run(args):
    return run_direct(args.index, args.input, args.bed, args.r)


def run_direct(index_filename, input_filenames, bed_filename=None, region=None):
    samples = [Sample(filename) for filename in input_filenames]

    index = h5py.File(index_filename, "r")

    print("#chrom", "start", "stop", sep="\t", end="\t")
    print(*samples, sep="\t")

    dmrs = defaultdict(list)

    if region:
        if region.find(":") == -1:
            chrom = region
            start, stop = 0, 99999999999
        else:
            chrom, start_stop = region.split(":")
            start, stop = map(int, start_stop.split("-"))
        dmrs[chrom].append((start, stop))
    elif bed_filename:
        for line in open(bed_filename, "r"):
            line = line.strip()
            if len(line) < 2: #skip empty lines
                continue

            if line.startswith("#"): #skip comments
                continue

            split = line.split("\t")
            if len(split) < 3:
                continue

            chrom, start, stop = split[:3]
            if chrom.startswith("chr"):
                chrom = chrom[3:]
            start = int(start)
            stop = int(stop)

            dmrs[chrom].append((start, stop))
    else:
        for chrom in index:
            stop = index[chrom]["cpg_positions"][-1]
            dmrs[chrom].append((1, stop))


    for chrom in natsorted(dmrs):
        if chrom not in samples[0].chromosomes():
            print("WARNING: {chrom} not in camel file".format(chrom=chrom), file=sys.stderr)
            continue
        indices = np.searchsorted(index[chrom]["cpg_positions"][:], dmrs[chrom])
        m_values = np.array([s.m_values(chrom) for s in samples]).T

        for (start, stop), (start_index, stop_index) in zip(dmrs[chrom], indices):
            print(chrom, start, stop, sep="\t", end="\t")
            print(*(np.sum(m_values[start_index:stop_index], axis=0) / (stop_index-start_index)))
#            averages = [round(np.average(s.m_values(chrom, start_index, stop_index)),2) for s in samples]
#            print(*averages, sep="\t")


def sub(parser):
    parser.add_argument('index', help='the camel index')
    parser.add_argument('input', nargs='+', help='the samples of which to build the average of')
    parser.add_argument('--bed', help='a bed file with regions')
    parser.add_argument('--r', help='a region in the form of chr:start-stop')


def main():
    parser = argparse.ArgumentParser()
    sub(parser)
    args = parser.parse_args()
    run(args)


if __name__ == "__main__":
    main()



import argparse
import h5py
import numpy as np
import sys
from collections import namedtuple

Location = namedtuple("Location", "chrom, start, stop")

def parse_region(region):
    split = region.split(':')
    if len(split) == 1:
        chrom, = split
        return Location(chrom, -1, -1)
    chrom, positions = split
    split = positions.split('-')

    if len(split) == 1:
        start, = split
        return Location(chrom, int(start), -1)

    start, stop = split
    return Location(chrom, int(start), int(stop))

def out_native(chrom, methylation, positions):
    meth = methylation[:]
    stack = np.hstack((positions, meth))
    fmt = "{chrom}\t%i\t%i\t%i\t%i\t%i".format(chrom=chrom)
    return (stack, fmt)


def out_merged(chrom, methylation, positions):
    meth = methylation[:]
    cov = np.sum(meth, axis=1).reshape(-1,1)

    m = np.sum(meth[:,(0,2)], axis=1).reshape(-1,1)

    stack = np.hstack((positions, m, cov))


    fmt = "{chrom}\t%i\t%i\t%i".format(chrom=chrom)
    return (stack, fmt)


def out_simple(chrom, methylation, positions):
    meth = methylation[:]
    cov = np.sum(meth, axis=1)
    cov_plus = np.copy(cov)
    cov_plus[cov_plus==0] = 1

    m = np.sum(meth[:,(0,2)], axis=1)
    rel = np.round((m*100 / cov_plus)).astype(np.int32).reshape((-1,1))
    stack = np.hstack((positions, rel))
    fmt = "{chrom}\t%i\t%i".format(chrom=chrom)
    return (stack, fmt)


def out_epp(chrom, methylation, positions):
    plus_minus = np.empty(shape=(len(methylation[:]) * 2), dtype=np.int32)
    plus_minus[::2] = ord("+")
    plus_minus[1::2] = ord("-")

    positions_doubled = np.hstack((positions - 1, positions - 1)).flatten()

    cov = np.sum(methylation[:].reshape((-1,2)), axis=1)
    cov_plus = np.copy(cov)
    cov_plus[cov_plus==0] = 1

    meth = methylation[:].reshape((-1,2))[:,0]

    rel = (meth*1000 / cov_plus).astype(np.int32)
    stack = np.vstack((positions_doubled, positions_doubled + 2, meth, cov, rel, plus_minus)).T
    fmt = "chr{chrom}\t%i\t%i\t'%i/%i'\t%i\t%c".format(chrom=chrom)

    return (stack, fmt)


def out_bismark_cov(chrom, methylation, positions):
    positions_doubled = np.hstack((positions, positions+1)).flatten()

    cov = np.sum(methylation[:].reshape((-1,2)), axis=1)
    cov_plus = np.copy(cov)
    cov_plus[cov_plus==0] = 1

    meth = methylation[:].reshape((-1,2))[:,0]
    umeth = methylation[:].reshape((-1,2))[:,1]

    rel = np.round((meth*100 / cov_plus)).astype(np.int32)
    stack = np.vstack((positions_doubled, positions_doubled, rel, meth, umeth)).T
    fmt = "chr{chrom}\t%i\t%i\t%i\t%i\t%i".format(chrom=chrom)
    return (stack, fmt)


def out_bed(chrom, methylation, positions):
    meth = methylation[:]
    cov = np.sum(meth, axis=1)
    cov_plus = np.copy(cov)
    cov_plus[cov_plus==0] = 1

    m = np.sum(meth[:,(0,2)], axis=1)
    rel = np.round((m*100 / cov_plus)).astype(np.int32).reshape((-1,1))
    stack = np.hstack((positions, positions+1, rel))
    fmt = "{chrom}\t%i\t%i\t%i".format(chrom=chrom)
    return (stack, fmt)


def out_granges(chrom, methylation, positions):
    meth = methylation[:]
    cov = np.sum(meth, axis=1).reshape((-1,1))
    m = np.sum(meth[:,(0,2)], axis=1).reshape((-1,1))

    stack = np.hstack((positions, positions, cov, m))

    sys.stdout.buffer.write(b"\t".join([b"seqnames", b"start", b"end", b"width", b"strand", b"T", b"M"])) #header
    sys.stdout.buffer.write(b"\n") #header
    fmt = "chr{chrom}\t%i\t%i\t1\t*\t%i\t%i".format(chrom=chrom)
    return (stack, fmt)


def __data__(samples, index, region=None, style="native"):
    samples = h5py.File(samples, 'r')
    index = h5py.File(index, 'r')

    coverage = False

    if region:
        region = parse_region(args.region)
        chromosomes = [region.chrom]
    else:
        chromosomes = [i for i in samples if not i == "meta"] # TODO change structure


    for chrom in chromosomes:
#        coverages = np.sum(samples[chrom]['methylation'][:], axis=1)
#        fix_coverages = coverages[:]
#        fix_coverages[coverages == 0] = 1 # fix coverage = 0
#        meth = np.sum(samples[chrom]['methylation'][:,(0,2)], axis=1)
#        ratio = meth * 100 / fix_coverages

        cpgs = index[chrom]["cpg_positions"]

        if region: 
            start_index = np.searchsorted(cpgs, region.start) if region.start > -1 else 0
            stop_index = np.searchsorted(cpgs, region.stop) + 1 if region.stop > -1 else -1 # TODO: +1 includes the last position if it equals a cpg position

            positions = cpgs[start_index:stop_index].reshape((-1,1))
            methylation = samples[chrom]["methylation"][start_index:stop_index]
        else:
            positions = cpgs[:].reshape((-1,1))
            methylation = samples[chrom]["methylation"]

        # choose output style
        if style == "native":
            out_func = out_native
        if style == "merged":
            out_func = out_merged
        elif style == "simple":
            out_func = out_simple
        elif style == "epp":
            out_func = out_epp
        elif style == "bismarkCov":
            out_func = out_bismark_cov
        elif style == "granges":
            out_func = out_granges
        elif style == "bed":
            out_func = out_bed

        stack, fmt = out_func(chrom, methylation, positions)
        yield chrom, stack, fmt


def data_dict(samples, index, region=None):
    for chrom, stack, fmt in __data__(samples, index, region, "merged"):
        for x in stack:
            yield {
                "chrom": chrom,
                "pos": x[0],
                "methylation": x[1],
                "coverage": x[2],
            }


def run(args):
    for chrom, stack, fmt in __data__(args.input, args.index, args.region, args.style):
        try:
            np.savetxt(sys.stdout.buffer, stack, fmt=fmt)
        except IOError:
            break


def sub(p):
    p.add_argument('input', help='the hdf5 file')
    p.add_argument('index',help='the index file')
    p.add_argument('region', nargs='?', help='a region to output', default=None)
#    p.add_argument('--epp', action='store_true', default=False)

    p.add_argument('--style', '-s', choices=["native", "merged", "simple", "epp", "bismarkCov", "granges", "bed"], default="native")
    p.add_argument('-c', action='store_true', default=False)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    sub(parser)
    args = parser.parse_args()
    run(args)



import math
from bitarray import bitarray

def calculate_bloomfilter_size(sample_size, error_rate):
    n = sample_size
    p = error_rate
    m = math.ceil((n*math.log(p)) / math.log(1.0 / (math.pow(2.0, math.log(2.0)))))
    k = round(math.log(2.0) * m / n)
    return m, k


class Bloomfilter:
    def __init__(self, size, hash_count):
        self.size = size
        self.hash_count = hash_count
        self.bit_array = bitarray(size)
        self.bit_array.setall(0)


    def add(self, t):
        for seed in range(self.hash_count):
            result = hash((seed, t)) % self.size
            self.bit_array[result] = 1


    def lookup(self, t):
        for seed in range(self.hash_count):
            result = hash((seed, t)) % self.size
            if self.bit_array[result] == 0:
                return False
        return True



import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import os, inspect
import matplotlib.gridspec as gridspec
import scipy.spatial.distance as distance
import scipy.cluster.hierarchy as sch
import numpy as np
import argparse

from functools import reduce
from camel.wrap.sample import Sample


def gen_data(samples, size=1000):
    sample_chromosomes = set(samples[0].chromosomes())
    chrom = [c for c in map(str, range(1,100)) if c in sample_chromosomes]

    # load coverages and methylation
    coverages = []
    M = [] 
    for s in samples:
        coverages.append(np.hstack([s.coverage(c) for c in chrom]))
        M.append(np.hstack([s.m_values(c, smooth_window=5) for c in chrom]))

    M = np.array(M).T
#    for c in coverages:
#        c[c==0] = -1

    valid = reduce(np.logical_and, [c > 10 for c in coverages])

    sd = np.std(M, axis=1) * valid
    sorted_indices = np.argsort(sd)[::-1]
    print(sorted_indices[:size])

    print(M)

    return M[sorted_indices[:size]]


# helper for cleaning up axes by removing ticks, tick labels, frame, etc.
def clean_axis(ax):
    """Remove ticks, tick labels, and frame from axis"""
    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_ticks([])
    for sp in ax.spines.values():
        sp.set_visible(False)


def plot(samples, labels=None):
    '''return fig fig of the clustering of samples'''
    # create a new figure
    fig = plt.figure(figsize=(10, 10))

    # use the top 1000 cpgs
    SIZE = 1000

    # permute test data and make dataframe
    data = gen_data(samples, SIZE)

    # reordering rows #
    pairwise_dists = distance.squareform(distance.pdist(data))

    # build the clustering
    clusters = sch.linkage(pairwise_dists,method='complete')
    dendogram = sch.dendrogram(clusters,color_threshold=np.inf,no_plot=True)

    data = data[dendogram['leaves']]

    #calculating pairwise distance matrix
    pairwise_dists = distance.squareform(distance.pdist(data.T, 'cityblock')) / SIZE
    clusters = sch.linkage(pairwise_dists,method='complete')


    #### init grid ####
    heatmap_grid = gridspec.GridSpec(2,1,wspace=0.0,hspace=0.01,height_ratios=[0.25,1])

    column_grid = gridspec.GridSpecFromSubplotSpec(2,1,subplot_spec=heatmap_grid[0,0],wspace=0.0,hspace=0.1,height_ratios=[1,0.15])


    #### dendogram ####
    sch.set_link_color_palette(['black'])
    dendogram_ax = fig.add_subplot(column_grid[0,0])
    dendogram = sch.dendrogram(clusters,color_threshold=np.inf)
    clean_axis(dendogram_ax)

    ### column colorbar ###
#    column_colorbar_ax = fig.add_subplot(column_grid[1,0])
#    column_colorbar_ax.imshow(np.array([[1] * len(samples1) + [2] * len(samples2)]),interpolation='nearest',aspect='auto',origin='lower', cmap="Set3")
#    clean_axis(column_colorbar_ax)


    #### heatmap ####
    heatmap_ax = fig.add_subplot(heatmap_grid[1,0])
    #ax = fig.add_axes([0.25, 0.25, 0.5, 0.5])
    heatmap_ax.imshow(data.T[dendogram['leaves']].T, interpolation='nearest',aspect='auto',origin='lower', cmap="RdBu_r")
    axes = heatmap_ax.get_axes()
    clean_axis(axes)


    #### col labels ####
    if not labels:
        labels = [s.basename for s in samples]

    labels = np.array(labels)
    heatmap_ax.set_xticks(np.arange(len(labels)))
    xlabelsL = heatmap_ax.set_xticklabels(labels[dendogram['leaves']])

    # rotate labels 90 degrees
    for label in xlabelsL:
        label.set_rotation(90)
    # remove the tick lines
    for l in heatmap_ax.get_xticklines() + heatmap_ax.get_yticklines(): 
        l.set_markersize(0)

    return fig


def run(args):
    samples = [Sample(s) for s in args.samples]
    fig = plot(samples, labels=args.labels)
    fig.savefig(args.output, bbox_inches='tight')


def sub(parser):
    parser.add_argument("output", help="the output file for the plot")
    parser.add_argument("--samples", nargs="+", help="methylation values in camel format")
    parser.add_argument("--labels", nargs="*", help="the labels of the samples for the clustering plot")


def main():
    parser = argparse.ArgumentParser(description='create a clustering plot')
    sub(parser)
    args = parser.parse_args()
    run(args)


if __name__ == "__main__":
    main()



import h5py
import numpy as np
import argparse
import sys

def sub(parser):
    parser.add_argument('input', nargs='+', help='the samples of which to build the average of')
    parser.add_argument('output', help='to calculated average methylation values')


def run(args):
    files = [h5py.File(f, "r") for f in args.input]
    out_file = h5py.File(args.output, "w")

    chromosomes = [c for c in files[0]]

    print("building average sample", file=sys.stderr)
    for chrom in chromosomes:
        print("chromosome", chrom, file=sys.stderr)
        coverages = np.sum([np.sum(f[chrom]['methylation'][:], axis=1) > 0 for f in files], axis=0)
        coverages[coverages == 0] = 1
        sum_values = np.sum([f[chrom]['methylation'][:] for f in files], axis=0)
        mean_values = sum_values / coverages[:,None]
        grp = out_file.create_group(chrom)
        dset = grp.create_dataset("methylation", dtype='i', data=mean_values, maxshape=mean_values.shape)


#    s = np.sum([f[chrom]['methylation'] for f in files])
#    print(s)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    sub(parser)
    args = parser.parse_args()
    mean_sample(args)



#calculates statistics
import numpy as np
from camel.wrap.sample import Sample

def cast_none(x, type=float):
    if x is None or np.isnan(x):
        x = -1
    return type(x)


def data(filenames):
    for i in filenames:
        sample = Sample(i)
        yield {
            "name" : sample.basename,
            "lanes": cast_none(sample.lane_count, int),
            "duplication_rate": cast_none(sample.duplication_rate),
            "efficiency": cast_none(sample.efficiency),
            "coverage": cast_none(sample.read_coverage),
            "convertion_rate": cast_none(sample.convertion_rate),
            "methylation": cast_none(sample.all_meth),
        }


def out(data):
    #formated print out the data
    print("Sample", "#Track", "Duplication Rate", "Efficiency", "Coverage", "Convertion", "Methylation", sep="\t")
    for d in data:
        print(d["name"], d["lanes"], d["duplication_rate"], d["coverage"], d["convertion_rate"], d["methylation"], sep="\t")


def run(args):
    out(data(args.input))


def sub(parser):
    parser.add_argument('input', nargs='+', help='the input camel call files')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='show statistics for camel call files')
    sub(parser)
    args = parser.parse_args()



#Access the meta information

import sys
import argparse
import numpy as np
import h5py

def isfloat(value):
  try:
    float(value)
    return True
  except ValueError:
    return False


def read(f, name):
    if name not in f["meta"]:
        return
    print(f["meta/" + name][0])


def delete(f, name):
    if name in f["meta"]:
        del f["meta/" + name]
    else:
        print("variable \"", name, "\" not found", sep ="", file=sys.stderr)


def write(f, name, value):
    if isfloat(value):
        value = float(value)
    else:
        value = np.string_(value)

    if not "meta" in f:
        f.create_group("meta")

    if not name in f["meta"]:
        f["meta"].create_dataset(name, data = [value], maxshape=(1,))
    else:
        f["meta"][name][0] = value


def show_all(f):
    for m in f["meta"]:
        print(m, "=", f["meta"][m][0])


def run(args):
    f = h5py.File(args.input) #open the hdf5 file

    if args.name == None:
        show_all(f)
        return

    if args.delete:
        delete(f, args.name)
        return 

    if args.s == None:
        read(f, args.name)
        return

    write(f, args.name, args.s)


def sub(parser):
    parser.add_argument('input', help='the input camel call file')
    parser.add_argument('name', nargs='?', help='the name of the meta information')
    parser.add_argument('-s', default=None, help='write this value as meta information')
    parser.add_argument('--delete', default=False, action='store_true', help='delete the meta information')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Access the meta information.')
    sub(parser)
    args = parser.parse_args()
    run(args)



'''converts a textfile to a camel-data-file'''

import h5py
import numpy as np
import argparse
import sys
import gzip

from camel.wrap.sample import Sample

def run(args):
    return run_direct(args.index, args.input, args.output, args.seperator, args.skiplines)


def write(index, chrom, positions, ms, covs, outfile):

    chrom_positions = index[chrom.decode()]["cpg_positions"][:]
    indices = np.searchsorted(chrom_positions, positions)
    chrom_pos_verify_values = chrom_positions[indices]
    valid = positions == chrom_pos_verify_values

    indices = indices[valid]
    positions = np.array(positions)[valid]
    ms = np.array(ms)[valid]
    covs = np.array(covs)[valid]

    final = np.zeros((len(chrom_positions), 4))
    final[:,0][indices] = ms
    final[:,1][indices] = covs - ms

    grp = outfile.create_group(chrom)
    dset = grp.create_dataset("methylation", dtype='i', data=final, maxshape=final.shape)


def run_direct(index_filename, input_filename, output_filenames, seperator="\t", skiplines=0):
    index = h5py.File(index_filename, "r")
    outfile = h5py.File(output_filenames, "w")

    curr_chrom = ""

    for i,line in enumerate(gzip.open(input_filename, "r")):
        if i < skiplines:
            continue

        split = line.strip().split(seperator.encode())
        chrom = split[0]

        if chrom.startswith(b"chr"):
            chrom = chrom[3:]

        if curr_chrom != chrom:
            if len(curr_chrom):
                write(index, curr_chrom, positions, ms, covs, outfile)
            positions = []
            ms = []
            covs = []
            curr_chrom = chrom
            print(curr_chrom.decode(), file=sys.stderr)

        positions.append(int(split[1])) #positions
        ms.append(int(split[2])) #m_values
        covs.append(int(split[3])) #coverages

    write(index, curr_chrom, positions, ms, covs, outfile)

    outfile.close()



def sub(parser):
    parser.add_argument('index', help='the camel index')
    parser.add_argument('input', help='the sample in text format')
    parser.add_argument('output', help='the filename of the resulting camel-data-file (.h5)')
    parser.add_argument('-s', "--seperator", default="\t", help='the seperator, that seperates the values')
    parser.add_argument('-l', "--skiplines", default=0, type=int, help="skip the first x lines")


def main():
    parser = argparse.ArgumentParser()
    sub(parser)
    args = parser.parse_args()
    run(args)


if __name__ == "__main__":
    main()



# coding: utf-8

import sys
from setuptools import setup

if sys.version_info < (3,3):
    print("At least Python 3.3 is required for camel.", file=sys.stderr)
    exit(1)
try:
    from setuptools import setup
except ImportError:
    print("Please install setuptools before installing camel.", file=sys.stderr)
    exit(1)

# set __version__ and DESCRIPTION
exec(open("camel/version.py").read())

setup(
    name='gi-camel',
    version=__version__,
    author='Christopher Schröder and Sven Rahmann',
    author_email='christopher.schroeder@tu-dortmund.de, sven.rahmann@uni-due.de',
    description=DESCRIPTION,
    zip_safe=False,
    license='MIT',
    url='https://bitbucket.org/christopherschroeder/camel',
    packages=['camel', 'camel.modules', 'camel.wrap', 'camel.helper'],
    entry_points={
        "console_scripts":
            ["camel = camel.camel:run",
            ]
        },
    package_data={'': ['*.css', '*.sh', '*.html']},
    install_requires=['numpy', 'scipy', 'seaborn', 'matplotlib'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ]
)




import argparse, re, sys
import struct
import h5py
import numpy as np


def sub(parser):
    parser.add_argument('fasta', help='the reference file in fasta format')
    parser.add_argument('--nome', action='store_true', help='do additional gcp indexing for nome-seq')
    parser.add_argument('output', help='the output index hdf5 file')


def run(args):
    print("Creating index for calling", file=sys.stderr)

    fasta, output = args.fasta, args.output
    nome = args.nome
    base_pos = 0
    chrom = ""

    fasta = open(fasta)
    index = h5py.File(output, 'w')

    line = fasta.readline()

    while True:
        if not line: #eof -> write file and end the loop
            write_hdf5(index, chrom, cpg_positions, gpc_positions, c_counts, g_counts, valid_cpg_c, valid_cpg_g, valid_gpc_g, valid_gpc_c, nome)
            break

        if line.startswith(">"): #new chrom
            if chrom:
                write_hdf5(index, chrom, cpg_positions, gpc_positions, c_counts, g_counts, valid_cpg_c, valid_cpg_g, valid_gpc_g, valid_gpc_c, nome)

            # clear the cpg_positions
            cpg_positions = [] # the cpg_positions
            gpc_positions = [] # the cpg_positions
            c_counts  = []
            g_counts  = []
            curr_c_count = 0
            curr_g_count = 0
            valid_cpg_c = []
            valid_cpg_g = []
            valid_gpc_g = []
            valid_gpc_c = []

            chrom = line.split()[0][1:] # set new chrom name
            print("Chromosome", chrom, file=sys.stderr) # output for debugging

            base_pos = 0 # begin with new base position for couting
            line = fasta.readline()
            continue

        # no new chrom
        # as long as there is an C or G at the end of the line
        # it is possible that the linebreak splits a CG
        nextline = fasta.readline()

        while (line.upper().endswith("C\n") or line.upper().endswith("G\n")) and not nextline.startswith(">"): # the \n garanties that still a line is available
            line = line.rstrip() + nextline #append next line
            nextline = fasta.readline()

        line = line.upper() # in the case the fasta file contains lower case bp

        last_start = 0
        for m in re.finditer("CG", line):
            cpg_positions.append(base_pos + m.start() + 1)
            curr_c_count += line.count("C", last_start, m.start() + 2)
            curr_g_count += line.count("G", last_start, m.start() + 2)
            c_counts.append(curr_c_count)
            g_counts.append(curr_g_count)
            last_start = m.start() + 2

            if nome:
                valid_cpg_g.append(m.start() == len(line) - 2 or not line[m.start() + 2] == "C") #CGC
                valid_cpg_c.append(m.start() == 0 or not line[m.start() -1] == "G") #GCG

        if nome:
            # GC indexing for nome-sequencing
            for m in re.finditer("GC", line):
                gpc_positions.append(base_pos + m.start() + 1)
                valid_gpc_c.append(m.start() == len(line) - 2 or not line[m.start() + 2] == "G") #GCG
                valid_gpc_g.append(m.start() == 0 or not line[m.start() -1] == "C") #CGC

        curr_c_count += line.count("C", last_start, len(line))
        curr_g_count += line.count("G", last_start, len(line))

        base_pos += len(line) - 1 # adjust base position
        line = nextline

    fasta.close()
    index.close()


def write_hdf5(index, chrom, cpg_positions, gpc_positions, c_counts, g_counts, valid_cpg_c, valid_cpg_g, valid_gpc_g, valid_gpc_c, nome=False,):
    gc_counts = np.transpose(np.array([c_counts, g_counts]))
    grp_chrom = index.create_group(chrom)
    grp_chrom.create_dataset("cpg_positions", dtype='i', data=np.array(cpg_positions), maxshape=(len(cpg_positions),))
    grp_chrom.create_dataset("gc_counts", dtype='i', data=gc_counts, maxshape=gc_counts.shape)
    if nome:
        grp_chrom.create_dataset("gpc_positions", dtype='i', data=np.array(gpc_positions), maxshape=(len(gpc_positions),))
        valid_cpg_c = np.array(valid_cpg_c, dtype=np.bool_)
        valid_cpg_g = np.array(valid_cpg_g, dtype=np.bool_)
        valid_cpg = np.vstack((valid_cpg_c,valid_cpg_g)).T

        valid_gpc_c = np.array(valid_gpc_c, dtype=np.bool_)
        valid_gpc_g = np.array(valid_gpc_g, dtype=np.bool_)
        valid_gpc = np.vstack((valid_gpc_g, valid_gpc_c)).T

        grp_chrom.create_dataset("valid_cpg", dtype=np.bool_, data=valid_cpg, maxshape=valid_cpg.shape)
        grp_chrom.create_dataset("valid_gpc", dtype=np.bool_, data=valid_gpc, maxshape=valid_gpc.shape)


def main():
    parser = argparse.ArgumentParser()
    sub(parser)
    args = parser.parse_args()
    run(args)

if __name__ == '__main__':
    main()
